/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"

void usage(int status)
{
	if (status !=0)
		fprintf (stderr, "Tente `%s --help' para mais informaes.\n",PROGRAM_NAME);
	else {
		fprintf(stdout,"\
Uso: %s ENDEREO [ARQUIVO]\n\
 ou: %s OPES\n\
\n\
OPES\n\
\n\
     -h, --help        mostra estas informaes e sai\n\
     -V, --version     mostra a verso do programa e sai\n\
\n\
ENDEREO\n\
\n\
     Linha   Coluna    Significado\n\
     -----------------------------------------------\n\
       N       [N]     N\n\
     N;M;I   [N,M,I]   N, M e I\n\
      N:M     [N:M]    trecho de N at M\n\
      N~M     [N~M]    a partir de N, ande de M em M\n\
\n\
Se o ARQUIVO for '-', l o texto da entrada padro (STDIN).\n",PROGRAM_NAME,PROGRAM_NAME);
fprintf(stdout,"Se o nmero do endereo for negativo, conta a partir do fim.\n\
Envie relatrios de erros para: %s\n",BUG_ADDRESS);
	}
	exit (status == 0 ? 0 : 1);
}

void get_file(struct struct_rac_data *p_rac_data)
{
	int c;
	long int i=0;
	long int allocated_num_line=LINE_NUM;
	long int actual_line_size=0;
	/* alloc vector to number of chars per line */
	if (!(p_rac_data->num_chars_line = (long int *) malloc(LINE_NUM * sizeof(long int)))){
		fprintf(stderr, "Error: alloc vector to number of chars per line\n");
			exit(1);
	}
	/* alloc numer of lines' memory */
	if (!(p_rac_data->line = (char **) malloc(LINE_NUM * sizeof(char *)))){
		fprintf(stderr, "Error: alloc number of lines\n");
		exit(1);
	}

	while (1) {
		i++;
		c = getc(p_rac_data->stream);
		/* size of line is enough ?! */
		if (actual_line_size - 1 <= i) {
			if (!actual_line_size) p_rac_data->line[p_rac_data->num_line] = NULL;
#ifdef DEBUG2
			printf("-- alloc size of line. line number = %li, size line = %li\n",p_rac_data->num_line,actual_line_size);

#endif
			actual_line_size += LINE_SIZE;
			if (actual_line_size >= LONG_MAX) {
				fprintf (stderr, "Error: overflow actual_line_size\n");
				exit(1);
			}
			p_rac_data->line[p_rac_data->num_line] = (char *) realloc (p_rac_data->line[p_rac_data->num_line], actual_line_size * sizeof (char));
			if (!p_rac_data->line[p_rac_data->num_line]) {
					fprintf (stderr, "Error: realloc line's memory\n");
					exit(1);
			}
		}
		p_rac_data->line[p_rac_data->num_line][i] = c;
		if (c == '\n' || c == EOF || c == '\r') {
			/* is this a DOS file ??*/
			if (c == '\r') 	{ 
				c = getc(p_rac_data->stream);
				if (c != '\n') {
						fprintf(stderr,"Error: unknown file\n");
						exit(1);
				}
			}
			p_rac_data->line[p_rac_data->num_line][i+1] = '\0';
			i--;
			if (c == EOF) 
					p_rac_data->line[p_rac_data->num_line][i+1] = '\0';
			p_rac_data->num_chars_line[p_rac_data->num_line] = i;

			if (c == EOF) {
				if (i == 0) p_rac_data->num_line--;
				break;
			}
			else {
				i = 0;
				actual_line_size = 0;
				/* dont need check. we check the allocated_num_line */
				p_rac_data->num_line++;
				/* number of lines are enough ?! */
				if (p_rac_data->num_line >= allocated_num_line - 1) {
#ifdef DEBUG2
					printf("-- realloc number of lines. line number = %li\n",p_rac_data->num_line);
#endif
					allocated_num_line += LINE_NUM;
					if (allocated_num_line >= LONG_MAX) {
						fprintf (stderr, "Error: overflow allocated_num_line\n");
						exit(1);
					}
					p_rac_data->line = (char **) realloc (p_rac_data->line, allocated_num_line * sizeof (char *));
					if (!p_rac_data->line) {
						fprintf (stderr, "Error: realloc num line's memory\n");
						exit(1);
					}
					p_rac_data->num_chars_line = (long int *) realloc(p_rac_data->num_chars_line, allocated_num_line * sizeof(long int));
					if (!p_rac_data->num_chars_line) {
						fprintf(stderr, "Error: realloc vector to number of chars per line\n");
						exit(1);
					}
				}
			}
		}
	}
}

/*
 * TODO: Merge the check_col and check_lines functions.
 *       Just one check and filter!!!
 * */
void check_col(long int *line, long int *col,struct struct_rac_data *p_rac_data)
{
	if (*col < 0) 
		*col = p_rac_data->num_chars_line[*line] + (*col+1);

	if (*col < 0) *col = 1;
	if (*col > p_rac_data->num_chars_line[*line]) *col = 0;
}

int check_lines(struct struct_rac_data *p_rac_data)
{
	if (p_rac_data->line_start > p_rac_data->num_line && p_rac_data->line_end < 0) return -1;	
	if (p_rac_data->line_start > p_rac_data->num_line && p_rac_data->line_end >= p_rac_data->num_line) return -1;	
	
	/* check. we cant overflow the limits */
	if (p_rac_data->line_start < 0) p_rac_data->line_start = p_rac_data->num_line + (p_rac_data->line_start+1);
	if (p_rac_data->line_end   < 0) p_rac_data->line_end = p_rac_data->num_line + (p_rac_data->line_end+1);

	if (p_rac_data->line_start == p_rac_data->line_end)
			if (p_rac_data->line_end < 1 || p_rac_data->line_end > p_rac_data->num_line)
					return -1;

	if (p_rac_data->line_start >= 1 && p_rac_data->line_start <= p_rac_data->num_line) {
			if (p_rac_data->line_end < 1) p_rac_data->line_end = 1;
			if (p_rac_data->line_end > p_rac_data->num_line) p_rac_data->line_end = p_rac_data->num_line;
	}
		
	if (p_rac_data->line_end <= p_rac_data->num_line && p_rac_data->line_end >= 1) {
			if (p_rac_data->line_start < 1) p_rac_data->line_start = 1;
			if (p_rac_data->line_start > p_rac_data->num_line) p_rac_data->line_start = p_rac_data->num_line;
	}

#ifdef DEBUG1
	printf("-- Definitive Range = from %li to %li\n",p_rac_data->line_start,p_rac_data->line_end);
#endif
	
	/* to guarantee that we do not overflow the limits of matrix */
	if (p_rac_data->line_start < 1 || p_rac_data->line_start > p_rac_data->num_line) return -1;
	if (p_rac_data->line_end < 1 || p_rac_data->line_end > p_rac_data->num_line) return -1;

	return 0;
}

void get_type(char *command,struct struct_rac_data *p_rac_data)
{
	char c;
	int i=0;

	for (i=0;i<=strlen(command);i++) {
		c = command[i];
		if (c == ':')
			p_rac_data->TYPE = TYPE_RANGE_LINE;
		else if (c == '[') {
			while (c != ']' && c){
				i++;
				c = command[i];
			}
		}
	}
	if (p_rac_data->TYPE == -1) p_rac_data->TYPE = 1;
}

/* We know. This function is really horrible. 
 * TODO: please improve!!!  */
void exec_range_lines(char *command,struct struct_rac_data *p_rac_data)
{
	long int i, j, start, end;
	
	if (p_rac_data->line_start < 0 && p_rac_data->line_end < 0)
		p_rac_data->line_end = p_rac_data->num_line + (p_rac_data->line_end+1);
	if (p_rac_data->line_start > 0 && p_rac_data->line_end + p_rac_data->num_line < 0) return;
	if (p_rac_data->line_start < 0 && p_rac_data->line_end > p_rac_data->num_line) return;
	if (p_rac_data->line_start + p_rac_data->num_line < 0 && p_rac_data->line_end <= 1) return;
	if (check_lines((struct struct_rac_data *)p_rac_data) != 0) return;

	if (p_rac_data->line_start <= p_rac_data->line_end) {
		for (i=p_rac_data->line_start;i<=p_rac_data->line_end;i=i+p_rac_data->STEP_LINE) {
			start = p_rac_data->col_start[0] == '\0' ? 1 : atoi(p_rac_data->col_start); 
			end = p_rac_data->col_end[0] == '\0' ? -1 : atoi(p_rac_data->col_end);

			check_col(&i,&start,(struct struct_rac_data *)p_rac_data);
			check_col(&i,&end,(struct struct_rac_data *)p_rac_data);

			if (i == p_rac_data->line_start) {
				for (j=start;j<=p_rac_data->num_chars_line[i];j++)
					if (start != 0)
						printf("%c",p_rac_data->line[i][j]);
			}
			else if (i == p_rac_data->line_end) {
				if (end == 0) end =p_rac_data->num_chars_line[i];
					for (j=1;j<=end;j++)
						printf("%c",p_rac_data->line[i][j]);
			}
			else {
				for (j=1;j<=p_rac_data->num_chars_line[i];j+=p_rac_data->STEP_LINE)
					printf("%c",p_rac_data->line[i][j]);
			}
			printf("\n");
		}
	}
	else {
		p_rac_data->STEP_LINE = -1;
		for (i=p_rac_data->line_start;i>=p_rac_data->line_end;i=i+p_rac_data->STEP_LINE) { 
			start = p_rac_data->col_start[0] == '\0' ? 1 : atoi(p_rac_data->col_start); 
			end = p_rac_data->col_end[0] == '\0' ? -1 : atoi(p_rac_data->col_end);

			check_col(&i,&start,(struct struct_rac_data *)p_rac_data);
			check_col(&i,&end,(struct struct_rac_data *)p_rac_data);
			if (i == p_rac_data->line_start) {
				for (j=start;j<=p_rac_data->num_chars_line[i];j++)
					printf("%c",p_rac_data->line[i][j]);
			}
			else if (i == p_rac_data->line_end) {
				for (j=1;j<=end;j++)
					printf("%c",p_rac_data->line[i][j]);
			}
			else {
				for (j=1;j<=p_rac_data->num_chars_line[i];j++)
					printf("%c",p_rac_data->line[i][j]);
			}
			printf("\n");
		}
	}
}

/*
 * TODO: This function too.
 *       needs improvements 
 */
void exec_range_cols(char *command,struct struct_rac_data *p_rac_data)
{
	long int i, j, start, end;
	int all_line;
	char *temp, *token, *col_temp1, *col_temp2, *col_STEP;

	if (p_rac_data->col_start[0] == '\0' && p_rac_data->col_end[0] != '\0')
			strncpy(p_rac_data->col_start,p_rac_data->col_end,strlen(p_rac_data->col_end));

	if (p_rac_data->line_start + p_rac_data->num_line < 0 && p_rac_data->line_end <= 1 && p_rac_data->STEP_LINE < 0) return;
	if (check_lines((struct struct_rac_data *)p_rac_data) != 0) return;

	i = p_rac_data->line_start;

	if (p_rac_data->line_start > p_rac_data->line_end) {
			j = p_rac_data->line_start;
			p_rac_data->line_start = p_rac_data->line_end;
			p_rac_data->line_end = j;
	}
	
	while (i >= p_rac_data->line_start && i <= p_rac_data->line_end) {
		temp = (char *) malloc(strlen(p_rac_data->col_start)+1);
		temp[0]='\0';
		all_line = 0;
		strcpy(temp,p_rac_data->col_start);
		token = strtok(temp, ",");
		p_rac_data->STEP_COL=1;
		while (token != NULL) {
			all_line = 1;
			col_temp1 = (char *) malloc (strlen(token)+1);
			col_temp1[0] = '\0';
			strncpy(col_temp1,token,strlen(token));
			token = strtok(NULL, ",");
			col_temp2 = strchr(col_temp1,':');
			col_STEP  = strchr(col_temp1,'~');
			/* is it a range ? */
			if (col_temp2 != NULL || col_STEP != NULL) {
				if (col_temp2 != NULL) {
					if (col_temp1[0] == ':') start = 1;
					else start = atoi(col_temp1);
				
					*col_temp2 = 0;
					col_temp2++;
			
					if (atoi(col_temp2) == 0)
						end = -1;
					else {
						end = atoi(col_temp2);
						if (start < 0 && end > p_rac_data->num_chars_line[i]) 
								continue;
						if (end > p_rac_data->num_chars_line[i])
							end = -1;
					}
				}	
				else if (col_STEP != NULL) {
						*col_STEP=0;
						col_STEP++;
						
						p_rac_data->STEP_COL = atoi(col_STEP);
						
						if (col_temp1[0] == ':') start = 1;
						else start = atoi(col_temp1);


						if (p_rac_data->STEP_COL > 0) end = -1;
						else if (p_rac_data->STEP_COL < 0) end = 1;
						else {
							fprintf(stderr,"Error: step col invalid\n");
							exit(1);
						}
				}
				if (start > p_rac_data->num_chars_line[i] && 
							p_rac_data->num_chars_line[i]+p_rac_data->STEP_COL < 0) continue;

				if (start > p_rac_data->num_chars_line[i] && end <= 0) continue;
				
				if (p_rac_data->num_chars_line[i] + start < 0 && p_rac_data->STEP_COL <=0)
						continue;
			
				if (start > 0 && end+p_rac_data->num_chars_line[i] < 0) continue;

				if (start+p_rac_data->num_chars_line[i] < 0 && end > 0) continue;
				
				check_col(&i,&start,(struct struct_rac_data *)p_rac_data);
				check_col(&i,&end,(struct struct_rac_data *)p_rac_data);
				if (start == 0) start = p_rac_data->num_chars_line[i];
				if (start == 1 && end <= 0) continue;

#ifdef DEBUG1
				printf("-- col start %li col end = %li\n",start,end);
#endif
				if (start <= end) {
					for (j=start;j<=end;j=j+p_rac_data->STEP_COL) {
						if (j < 0 ) goto out; 
						if (j > 0 && j <= p_rac_data->num_chars_line[i])
							printf("%c",p_rac_data->line[i][j]);
					}
				}
				else {
					if (p_rac_data->STEP_COL == 1) p_rac_data->STEP_COL = -1;
					for (j=start;j>=end;j=j+p_rac_data->STEP_COL)
						if (j > 0 && j <= p_rac_data->num_chars_line[i])
							printf("%c",p_rac_data->line[i][j]);
				}
			}
			else { 
				start = atoi(col_temp1);
				if (start < 0)
					start = start + p_rac_data->num_chars_line[i] + 1;
				if (start > 0 && start <= p_rac_data->num_chars_line[i])
					printf("%c",p_rac_data->line[i][start]);
			}
out:
			free(col_temp1);
		}
		/* print line*/
		if (all_line == 0) {
			for (j=1;j<=p_rac_data->num_chars_line[i];j++)
				printf("%c",p_rac_data->line[i][j]);
		}
		printf("\n");
		i = i + p_rac_data->STEP_LINE;
	}
}

void exec_cmd(char *command,struct struct_rac_data *p_rac_data)
{

#ifdef DEBUG1
	printf("-- p_rac_data->col_start = %s\n",p_rac_data->col_start);
	printf("-- p_rac_data->col_end   = %s\n",p_rac_data->col_end);
#endif

	if (p_rac_data->TYPE == TYPE_RANGE_LINE) {
#ifdef DEBUG1
		printf("-- p_rac_data->TYPE = TYPE_RANGE_LINE\n");
#endif
		exec_range_lines(command,(struct struct_rac_data *)p_rac_data);
	}
	
	if (p_rac_data->TYPE == TYPE_RANGE_COL) {
#ifdef DEBUG1
		printf("-- p_rac_data->TYPE = TYPE_RANGE_COL\n");
#endif
		exec_range_cols(command,(struct struct_rac_data *)p_rac_data);
	}
}


void get_ranges(char *command, struct struct_rac_data *p_rac_data)
{
	char *p1, c;
	p1 = (char *) malloc(strlen(command)+1);
	
	#ifdef DEBUG1
		printf("-- p_rac_data->col_start = %s\n",p_rac_data->col_start);
		printf("-- p_rac_data->col_end   = %s\n",p_rac_data->col_end);
	#endif
	
	free(p_rac_data->col_start);
	free(p_rac_data->col_end);

	p_rac_data->col_start = (char *) malloc(strlen(command)+1);
	p_rac_data->col_end   = (char *) malloc(strlen(command)+1);
	p1[0] = '\0';
	p_rac_data->col_start[0] = '\0';
	p_rac_data->col_end[0]   = '\0';
	strncpy(&c,command,1);
	while(c) {
		if (c == ':' || c == '~') {
			if (atoi(p1) == 0) 
				p_rac_data->line_start = 1;
			else 
				p_rac_data->line_start = atoi(p1);
			if (c == '~') {
				command++;
				strncpy(&c,command,1);
				p_rac_data->STEP_LINE = atoi(command);
				if (p_rac_data->STEP_LINE >= 0) p_rac_data->line_end = -1;
				else p_rac_data->line_end = 1;
				while (c != '[' && c) {
					command++;
					strncpy(&c,command,1);
				}
			}
			else
				command++;
			strncpy(&c,command,1);
			*p1 = 0;
			while(c) {
				if (c == '[') {
					while(c != ']' && c) {
						if (c != '[' && c != ']') 
							strncat(p_rac_data->col_end,command,1);
						command++;
						strncpy(&c,command,1);
					}
					if (c != ']') {
						fprintf(stderr, "Error: Invalid Range\n");
						exit(1);
					}
#ifdef DEBUG2
					printf("-- p_rac_data->col_end = %s\n",p_rac_data->col_end);
#endif
				}
				strncat(p1,command,1);
				command++;
				strncpy(&c,command,1);
			}
			if (p_rac_data->line_end != -1 && p_rac_data->line_end != 1) {
				if (atoi(p1) == 0) p_rac_data->line_end = p_rac_data->num_line;
				else p_rac_data->line_end = atoi(p1);
			}
		}
		if (c == '[') {
			p_rac_data->line_start = atoi(p1);
			while(c != ']' && c) {
				if (c != '[' && c != ']') strncat(p_rac_data->col_start,command,1);
				command++;
				strncpy(&c,command,1);
			}
			if (c != ']') {
				fprintf(stderr, "Error: Invalid Range\n");
				exit(1);
			}
#ifdef DEBUG2
			printf("-- p_rac_data->col_start = %s\n",p_rac_data->col_start);
#endif
		}
		strncat(p1,command,1);
		command++;
		strncpy(&c,command,1);
	}

	if (p_rac_data->col_start[0] == '\0' && p_rac_data->line_start == 0) {
			p_rac_data->line_start = atoi(p1);
			if (p_rac_data->col_end[0] == '\0' && p_rac_data->line_end == 0) 
				p_rac_data->line_end = atoi(p1);
	}
	if (p_rac_data->col_start[0] != '\0' && p_rac_data->line_start != 0 && p_rac_data->col_end[0] == '\0'  && p_rac_data->line_end == 0) 
			p_rac_data->line_end = p_rac_data->line_start;
	if (p_rac_data->col_start[0] != '\0' && p_rac_data->line_start == 0 && p_rac_data->col_end[0] == '\0'  && p_rac_data->line_end == 0) {
			p_rac_data->line_start = 1;
			p_rac_data->line_end   = p_rac_data->num_line;
	}
	free(p1);
#ifdef DEBUG1
	printf("-- p_rac_data->line_start = %li\n",p_rac_data->line_start);
	printf("-- p_rac_data->line_end   = %li\n",p_rac_data->line_end);
#endif
}

int main(int argc, char *argv[])
{
	long int i, j, k;
	int cmd_num, cmd_size;
	char *token, **command;

	struct struct_rac_data rac_data;

	bzero(&rac_data,sizeof(rac_data));
	rac_data.num_line=1;

	if (argc < 2)
		usage(1);

	if ((!strcmp(argv[1],"-h")) || (!strcmp(argv[1],"--help")))
		usage(0);
	
	if (!strcmp(argv[1],"-V") || !strcmp(argv[1],"--version")) {
		fprintf(stdout, "%s verso %s\n", PROGRAM_NAME,VERSION);
		exit(0);
	}

	if (argc == 2 || (!strcmp(argv[2],"-"))) {
#ifdef DEBUG1
		printf("-- Reading from stdin\n");
#endif
		rac_data.stream = stdin;
	}
	else {
		if (!(rac_data.stream = fopen(argv[2], "r"))) {
			perror(argv[2]);
			exit(1);
		}
	}

	/* put file into matrix */
	get_file(&rac_data);

#ifdef DEBUG1
	printf("-- Total number of lines %li\n",rac_data.num_line);
#endif

	if (!(command = (char **) malloc(CMD_SIZE * sizeof (char *)))){
		fprintf(stderr, "Error: alloc command's memory\n");
		exit(1);
	}

	cmd_num = CMD_SIZE;
	cmd_size = 0;
	
	/* Get command by command */
	token = strtok(argv[1], ";");
	k=0;
	while (token != NULL) {
		for (i=0, j=0;i<strlen(token);i++) {
			if (cmd_size == 0) 	command[k] = NULL;
			/* erase the blank characters */
			if (j >= cmd_size - 1) {
#ifdef DEBUG2
				printf("-- rac_data.line = %li, realloc size of commands = %d\n",k,cmd_size);
#endif
				cmd_size += CMD_SIZE;
				if (cmd_size >= INT_MAX) {
					fprintf (stderr, "Error: overflow cmd_size\n");
					exit(1);
				}
				command[k] = (char *) realloc (command[k], cmd_size * sizeof (char));
				if (!command[k]) {
					fprintf (stderr, "Error: realloc command's memory\n");
					exit(1);
				}
			}
			if (token[i] != ' ') command[k][j++] = token[i];
		}
		token = strtok(NULL, ";");
		command[k][j] = '\0';
		cmd_size = 0;
		k++;
		if (k >= cmd_num) {
#ifdef DEBUG2
			printf("-- realloc number of commands = %d\n",cmd_num);
#endif
			cmd_num += CMD_SIZE;
			if (cmd_num >= INT_MAX) {
				fprintf (stderr, "Error: overflow cmd_num\n");
				exit(1);
			}
			command = (char **) realloc(command, cmd_num * sizeof(char *));
			if (!command) {
				fprintf (stderr, "Error: realloc command's memory\n");
				exit(1);
			}
		}
	}
	for (i=0;i<k;i++) {
#ifdef DEBUG1
		printf("-- Command = %s\n",command[i]);
#endif
		rac_data.line_start = rac_data.line_end = 0;
		rac_data.STEP_LINE=1;
		rac_data.TYPE=-1;
		/* it executes the command */
		get_ranges(command[i],&rac_data);
		get_type(command[i],&rac_data);
		exec_cmd(command[i],&rac_data);
		free(command[i]);
	}

	for (i=0;i<rac_data.num_line;i++) free(rac_data.line[i]);
	free(rac_data.line);
	fclose(rac_data.stream);

	exit(0);
}


